data BinTree a = Leaf | Node { left :: BinTree a, right :: BinTree a, this :: a} deriving Show

newtype Myset a = Myset (BinTree a)

instance Show a => Show (Myset a)
 where show (Myset t) = if length (tree2list t)==0 then "{ }" else "{ " ++ foldl (\x y -> x++", "++ show y) (show $ head $ tree2list t) (tail $ tree2list t) ++ " }" 


tree2list :: BinTree a -> [a]
tree2list Leaf = []
tree2list (Node l r x) = (tree2list l)++(x:(tree2list r))

set2list :: Myset a -> [a]
set2list (Myset t) = tree2list t

list2set :: Ord a => [a] -> Myset a
list2set xs = foldr insert (Myset Leaf) xs

insertTree :: Ord a => a -> BinTree a -> BinTree a
insertTree x Leaf = Node Leaf Leaf x
insertTree x (Node l r y)
 | x==y = Node l r y
 | x<y = Node (insertTree x l) r y
 | otherwise = Node l (insertTree x r) y

insert :: Ord a => a -> Myset a -> Myset a
insert x (Myset t) = if (inTree x t) then (Myset t) else Myset (insertTree x t)

inTree :: Ord a => a -> BinTree a -> Bool
inTree _ Leaf = False
inTree x (Node l r y)
 | x==y = True
 | x<y  = inTree x l
 | otherwise = inTree x r

belongs :: Ord a => a -> Myset a -> Bool
belongs x (Myset t) = inTree x t

sizeTree :: BinTree a -> Int
sizeTree Leaf = 0
sizeTree (Node l r _) = 1 + sizeTree l + sizeTree r

size :: Myset a -> Int
size (Myset t) = sizeTree t

union :: Ord a => Myset a -> Myset a -> Myset a
union (Myset Leaf) s2 = s2
union (Myset (Node l r x)) s2 = (union (Myset r) (union (Myset l) (insert x  s2)))

intersection :: Ord a => Myset a -> Myset a -> Myset a
intersection (Myset Leaf) _ = Myset Leaf
intersection (Myset (Node l r x)) s2 = if belongs x s2 then insert x rest else rest 
  where
   rest = union (intersection (Myset l) s2) (intersection (Myset r) s2)

--Alternative implementation
intersection' :: Ord a => Myset a -> Myset a -> Myset a
intersection' (Myset Leaf) _ = Myset Leaf
intersection' (Myset (Node l r x)) t = list2set $ [ x | x<-set2list (Myset (Node l r x)), belongs x t]

difference :: Ord a => Myset a -> Myset a -> Myset a
difference (Myset Leaf) _ = Myset Leaf
difference (Myset (Node l r x)) s2 
  | not (belongs x s2) = Myset (Node l1 r1 x)
  | otherwise = union (Myset l1) (Myset r1)
 where 
  (Myset l1) = difference (Myset l) s2
  (Myset r1) = difference (Myset r) s2

subset :: Ord a => Myset a -> Myset a -> Bool
subset (Myset Leaf) _ = True
subset (Myset (Node l r x)) s2 = subset (Myset l) s2 && subset (Myset r) s2 && belongs x s2 

instance Ord a => Eq (Myset a) 
  where
   s1 == s2 = subset s1 s2 && subset s2 s1

height :: Myset a -> Int
height (Myset Leaf) = 0
height (Myset (Node l r _)) = 1 + max (height (Myset l)) (height (Myset r))

-- test: random array

myrand :: Integer -> Integer
myrand x = (134775813*x+1) `mod` (2^32)

randlist :: Int -> [Integer] -> [Integer]
randlist 0 xs = xs
randlist n (x:xs) = randlist (n-1) ((myrand x):x:xs)

-- verifications

longlist = randlist 100000 [1]
longlist2 = randlist 100000 [(longlist !! 10)]
