data BinTree a = Leaf | Node { left :: BinTree a, right :: BinTree a, this :: a, h :: Int} deriving Show

newtype Myset a = Myset (BinTree a)

instance Show a => Show (Myset a)
 where show (Myset t) = if length (tree2list t)==0 
                          then "{ }" 
                          else "{ " ++ foldr (\x y -> show x++", "++ y) (show $ last $ tree2list t) (init $ tree2list t) ++ " }" 

heightS :: Myset a -> Int
heightS (Myset s) = height s

height :: BinTree a -> Int
height Leaf = 0
height (Node _ _ _ h) = h

tree2list :: BinTree a -> [a]
tree2list Leaf = []
tree2list (Node l r x _) = (tree2list l)++(x:(tree2list r))

set2list :: Myset a -> [a]
set2list (Myset t) = tree2list t

list2set :: Ord a => [a] -> Myset a
list2set xs = foldr insert (Myset Leaf) xs

rotateLeft :: BinTree a -> BinTree a
rotateLeft Leaf = Leaf
rotateLeft (Node l Leaf x h) = Node l Leaf x h
rotateLeft (Node l (Node rl rr xr _) x _) = Node (Node l rl x h') rr xr (1+max h' (height rr)) where h'=1+max (height l) (height rl)

rotateRight :: BinTree a -> BinTree a
rotateRight Leaf = Leaf
rotateRight (Node Leaf r x h) = Node Leaf r x h
rotateRight (Node (Node ll lr xl _) r x _) = Node ll (Node lr r x h') xl (1+max h' (height ll)) where h'=1+max (height lr) (height r)

rebalance :: BinTree a -> BinTree a
-- assumption: root is only possibly unbalanced node
rebalance Leaf = Leaf
rebalance (Node l r x _)
 | abs (height l-height r)<2 = Node l r x (1+ max (height l) (height r))
rebalance (Node l r x h) --root is indeed unbalanced
 | height l > height r = if height ll>height lr then rotateRight (Node (rotateRight l) r x h) else rotateRight (Node (rotateLeft l) r x h)
 | otherwise = if height rl>height rr then rotateLeft (Node l (rotateRight r) x h) else rotateLeft (Node l (rotateLeft r) x h)
 where
  (Node ll lr xl _)=l
  (Node rl rr xr _)=r

insertTree :: Ord a => a -> BinTree a -> BinTree a
insertTree x Leaf = Node Leaf Leaf x 1
insertTree x (Node l r y h)
 | x==y = Node l r y h
 | x<y = rebalance $ Node (rebalance $ insertTree x l) r y h
 | otherwise = rebalance $ Node l (rebalance $ insertTree x r) y h

insert :: Ord a => a -> Myset a -> Myset a
insert x (Myset t) = if (inTree x t) then (Myset t) else Myset (insertTree x t)

inTree :: Ord a => a -> BinTree a -> Bool
inTree _ Leaf = False
inTree x (Node l r y _)
 | x==y = True
 | x<y  = inTree x l
 | otherwise = inTree x r

belongs :: Ord a => a -> Myset a -> Bool
belongs x (Myset t) = inTree x t

sizeTree :: BinTree a -> Int
sizeTree Leaf = 0
sizeTree (Node l r _ _) = 1 + sizeTree l + sizeTree r

size :: Myset a -> Int
size (Myset t) = sizeTree t

union :: Ord a => Myset a -> Myset a -> Myset a
union (Myset Leaf) s2 = s2
union (Myset (Node l r x _)) s2 = (union (Myset r) (union (Myset l) (insert x  s2)))

intersection :: Ord a => Myset a -> Myset a -> Myset a
intersection (Myset Leaf) _ = Myset Leaf
intersection (Myset (Node l r x _)) s2 = if belongs x s2 then Myset (Node l1 r1 x (1+max (height l1) (height r1))) else union (Myset l1) (Myset r1)
  where
   (Myset l1) = intersection (Myset l) s2
   (Myset r1) = intersection (Myset r) s2

difference :: Ord a => Myset a -> Myset a -> Myset a
difference (Myset Leaf) _ = Myset Leaf
difference (Myset (Node l r x _)) s2 
  | not (belongs x s2) = insert x base
  | otherwise = base
 where 
  (Myset l1) = difference (Myset l) s2
  (Myset r1) = difference (Myset r) s2
  base = union (Myset l1) (Myset r1)

subset :: Ord a => Myset a -> Myset a -> Bool
subset (Myset Leaf) _ = True
subset (Myset (Node l r x _)) s2 = subset (Myset l) s2 && subset (Myset r) s2 && belongs x s2 

instance Ord a => Eq (Myset a) 
  where
   s1 == s2 = subset s1 s2 && subset s2 s1

-- test: random array

myrand :: Integer -> Integer
myrand x = (134775813*x+1) `mod` (2^32)

randlist :: Int -> [Integer] -> [Integer]
randlist 0 xs = xs
randlist n (x:xs) = randlist (n-1) ((myrand x):x:xs)

-- verifications

longlist = randlist 100000 [1]
longlist2 = randlist 100000 [(longlist !! 10)]

